/*--------------------------------------------------------------------


spherical interpolation routine based on SSRFPACK (netlib.org) by Renka,
see the ssrfpack routines for comments and descriptions of parameters

this driver part of taken from "surface" as of GMT 3.4.4, original
copyright notice follows below

modifications by Thorsten Becker (thwbecker@post.harvard.edu)

WARNING: this "release" is experimental!

see file README.ssrf for some more instructions

$Id: ssrf_surface.c,v 1.18 2004/09/08 00:52:01 twb Exp twb $


original copyright notice:

*	Id: surface.c,v 1.4.4.7 2003/04/17 22:39:25 pwessel Exp 
*
*	Copyright (c) 1991-2003 by P. Wessel and W. H. F. Smith
*	See COPYING file for copying and redistribution conditions.
*
*	This program is free software; you can redistribute it and/or modify
*	it under the terms of the GNU General Public License as published by
*	the Free Software Foundation; version 2 of the License.
*
*	This program is distributed in the hope that it will be useful,
*	but WITHOUT ANY WARRANTY; without even the implied warranty of
*	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*	GNU General Public License for more details.
*
*	Contact info: gmt.soest.hawaii.edu


*--------------------------------------------------------------------*/
#include "ssrf_surface.h"


int main(int argc, char **argv)
{
  int	i, j, error = FALSE,n_files,nz,istart,istop,iinc;
  FILE	*fp_in = stdin;	/* File pointer  */  
  char	*grdfile, format[500],*layerfile=NULL,outname[500];
  my_boolean is_three_d=FALSE,my_verbose = FALSE, randomize = FALSE,
    check_for_duplicates = TRUE, invert_i_loop = TRUE;
  struct GRD_HEADER h;
  struct lay *layer;
  float *df=NULL,*fi=NULL;
  double x_min,y_min,x_max,y_max,yinc,xinc,x,y,xlim=0.0,ylim=0.0;
  struct ssrf_par *sp;  
  
  /* 
     initialize parameters for the SSRF package 
  */
  df = NULL;			/* has to be init as NULL! */
  init_sp_structure(&sp);
  /* 
     set default interpolation mode.  for specifics and defaults, see
     ssrf_driver routine
  */
  sp->imode = 2;		/* use the c-1, global gradient,
				   method */
  /* 
     initialize GMT stuff 
  */
  grdfile = 0;
  layerfile = 0;
  /* 
     initialize the GMT stuff 
  */
  argc = GMT_begin (argc, argv);
  GMT_grd_init (&h, argc, argv, FALSE);
  /* 

  */
  x_min = y_min = 0.0;
  x_max = y_max = 0.0;
  xinc = yinc = 0.0;
  /* 

  check for parameters
  
  */
  n_files =0;
  for (i = 1; i < argc; i++) {
    if (argv[i][0] == '-') {
      switch (argv[i][1]) {
	
	/* Common parameters */
	
      case 'H':
      case 'R':
      case ':':
      case 'V':
      case '\0':
	error += GMT_get_common_args (argv[i], &x_min, &x_max, &y_min, &y_max);
	break;
	/* Supplemental parameters */
      case 'b':	/* Input triplets are binary, not ascii */
	error += GMT_io_selection (&argv[i][2]);
	break;
      case 'G':
	grdfile = &argv[i][2];
	break;
      case 'D':
	layerfile = &argv[i][2];
	is_three_d = TRUE;
	break;
      case 'S':
	randomize = TRUE;
	break;
      case 'C':
	check_for_duplicates = FALSE;
	break;
      case 'I':
	GMT_getinc (&argv[i][2], &xinc, &yinc);
	break;
      default:
	error = TRUE;
	GMT_default_error (argv[i][1]);
	break;
      }
    }
    else {
      if ((fp_in = GMT_fopen(argv[i], GMT_io.r_mode)) == NULL) {
	fprintf (stderr, "ssrf_surface: cannot open input data file %s\n", 
		 argv[i]);
	exit (EXIT_FAILURE);
      }
      n_files++;
    }
  }
  
  if (argc == 1 || GMT_quick) {	/* Display usage */
    fprintf (stderr, "ssrf_surface - fit regular grid to points on sphere using SSRFPACK\n");
    fprintf (stderr, "\n$Id: ssrf_surface.c,v 1.18 2004/09/08 00:52:01 twb Exp twb $\n\n");
    fprintf (stderr, "usage: ssrf_surface [xyz-file] -G<output_grdfile_name> -I<xinc>[m|c][/<yinc>[m|c]]\n");
    fprintf (stderr, "\t-R<west>/<east>/<south>/<north> [-D<layerfile>]\n");
    fprintf (stderr, "\t [-:] [-bi[s][<n>]]\n\n");
    
    if (GMT_quick) exit (EXIT_FAILURE);
    
    fprintf (stderr, "\tssrf_surface will read from standard input or a single <xyz-file>.\n\n");
    fprintf (stderr, "\tRequired arguments to ssrf_surface:\n");
    fprintf (stderr, "\t-G sets output grd File name (i.grd will be appended for layer i in case -D)\n");
    fprintf (stderr, "\t-I sets the Increment of the grid; enter xinc, optionally xinc/yinc.\n");
    fprintf (stderr, "\t\tDefault is yinc = xinc.  Append an m [or c] to xinc or yinc to indicate minutes [or seconds]\n");
    fprintf (stderr, "\t\te.g.  -I10m/5m grids longitude every 10 minutes, latitude every 5 minutes.\n\n");
    GMT_explain_option ('R');
    fprintf (stderr, "\n\tOPTIONS:\n");
    GMT_explain_option ('H');
    GMT_explain_option ('V');
    GMT_explain_option (':');
    GMT_explain_option ('i');
    GMT_explain_option ('n');
    fprintf (stderr, "\t\tDefault is 3 input columns (but see -D)\n\n");

    fprintf (stderr, "\t-C switches the check for duplicate input points OFF\n");
    fprintf (stderr, "\t   (on by default). This will speed up the code but will lead \n");
    fprintf (stderr, "\t   to errors in the triangulation if duplicates are read in.\n\n");

    fprintf (stderr, "\t-D sets the layer file name. In this case the data is lon lat z value\n");
    fprintf (stderr, "\t   and sorted into several grids (no vertical interpolation) according \n");
    fprintf (stderr, "\t   to the depths given in free format in the layer file.\n\n");
    
    fprintf (stderr, "\t-S reshuffles the input data in random order, recommended for some\n");
    fprintf (stderr, "\t   moderately regular (i.e. non-random) distributions (off by default)\n\n");
    GMT_explain_option ('.');
    exit (EXIT_FAILURE);
  }

  if (!project_info.region_supplied) {
    fprintf (stderr, "%s: GMT SYNTAX ERROR:  Must specify -R option\n", GMT_program);
    error++;
  }
  if (xinc <= 0.0 || yinc <= 0.0) {
    fprintf (stderr, "%s: GMT SYNTAX ERROR -I option.  Must specify positive increment(s)\n", GMT_program);
    error++;
  }
  if (!grdfile) {
    fprintf (stderr, "%s: GMT SYNTAX ERROR option -G:  Must specify output file\n", GMT_program);
    error++;
  }	
  if (GMT_io.binary[0] && gmtdefs.io_header) {
    fprintf (stderr, "%s: GMT SYNTAX ERROR.  Binary input data cannot have header -H\n", GMT_program);
    error++;
  }
  if (GMT_io.binary[0] && GMT_io.ncol[0] == 0) GMT_io.ncol[0] = 3;
  if (GMT_io.binary[0] && GMT_io.ncol[0] < 3) {
    fprintf (stderr, "%s: GMT SYNTAX ERROR.  Binary input data (-bi) must have at least 3 columns\n", GMT_program);
    error++;
  }
  if (n_files > 1) {
    fprintf (stderr, "%s: GMT SYNTAX ERROR.  Only one input file allowed.  Preprocess you files with blockmean/median/mode first\n", GMT_program);
    error++;
  }
  if (error) 
    exit (EXIT_FAILURE);
  GMT_put_history (argc, argv);	/* Update .gmtcommands */
  my_verbose = (gmtdefs.verbose)?(TRUE):(FALSE);
  if (GMT_io.binary[0] && gmtdefs.verbose) {
    char *type[2] = {"double", "single"};
    fprintf (stderr, "%s: Expects %d-column %s-precision binary data\n", 
	     GMT_program, GMT_io.ncol[0], type[GMT_io.single_precision[0]]);
  }
  /* set grid limits */
  h.x_min = x_min;
  h.x_max = x_max;
  h.y_min = y_min;
  h.y_max = y_max;
  h.x_inc = xinc;
  h.y_inc = yinc;

  GMT_grd_RI_verify (&h, 1);
  h.nx = irint ((x_max - x_min)/xinc) + 1;
  h.ny = irint ((y_max - y_min)/yinc) + 1;

  if (gmtdefs.verbose) {
    sprintf (format, "W: %s E: %s S: %s N: %s nx: %%d ny: %%d\n", gmtdefs.d_format, gmtdefs.d_format, gmtdefs.d_format, gmtdefs.d_format);
    fprintf (stderr, format, h.x_min, h.x_max, h.y_min, h.y_max, h.nx, h.ny);
  }
  /* 

  read in the data, for 2-D or 3-D

  */
  read_data(&layer,&nz,fp_in,is_three_d,layerfile);
  /* 
     interpolate on centers of blocks 
  */
  ylim = y_min - 1e-6;
  xlim = x_max + 1e-6;
  
  if(!invert_i_loop){		/* this was put in for debugging purposes  */
    istart = 0;   istop = nz;  iinc = 1;
  }else{
    istart = nz-1;istop = -1;   iinc= -1;
  }

  for(i=istart;i != istop;i += iinc){
    /* 
       loop through depth layers */
    /* 
       get  for the gridded interpolation 
    */
    if(i == istart)
      fi = (float *) GMT_memory (VNULL, (size_t)h.nx * h.ny, 
				 sizeof(float), GMT_program);
    /* 

       interpolate on the center of the grid points

    */
    /* 
       initialize the triangulation and remove duplicates from this 
       layer, if any
    */
    ssrf_driver(layer[i].lon, layer[i].lat, layer[i].f,
		&df,&layer[i].n,0,0,fi,sp,TRUE,randomize,
		check_for_duplicates,my_verbose);
    /* 
       generate a grid with interpolated value 
    */
    j = 0;
    for(y=y_max;y >= ylim;y -= yinc){
      for(x=x_min;x <= xlim;x += xinc){
	ssrf_driver(layer[i].lon, layer[i].lat, layer[i].f,
		    &df,&layer[i].n,x,y,(fi+j),sp,FALSE,
		    FALSE,FALSE,my_verbose);
	//	fprintf(stderr,"%g %g %g\n",x,y,fi[j]);	exit(-1);
	j++;
      }
    }
    /* 
       
    write GRD file
    
    */
    if(nz == 1){
      sprintf(h.remark,"generated by %s based on %i data points in single layer",
	      argv[0],layer[i].n);
      sprintf(outname,"%s",grdfile);
    }else{
      sprintf(h.remark,"based on %i points in layer %i out of %i at depth %g",
	      layer[i].n,i+1,nz,layer[i].z);
      sprintf(outname,"%s.%i.grd",grdfile,i+1);
    }
    if (GMT_write_grd (outname, &h, fi, 0.0, 0.0, 0.0, 0.0, 
		       GMT_pad, FALSE)) {
      fprintf (stderr, "%s: error writing file %s\n", 
	       GMT_program, outname);
      exit (EXIT_FAILURE);
    }
    if(gmtdefs.verbose)
      fprintf (stderr, "%s: written to %s\n", 
	       GMT_program, outname);
  } /* end i layer loop */
  for(i=0;i < nz;i++){
    free (layer[i].lon); 
    free (layer[i].lat); 
    free (layer[i].f); 
  } 
  free (fi);
  free (df);
  
  GMT_end (argc, argv);

  return 0;
}

/* 


read in 

lon lat value 

triplets from stream fp_in and store in the rlon,rlat,and f
arrays. returns the number of items as n

if is_three_d is set, will expect

lon lat z value 

values on roughly evenely spaced layers as specified in the layerfile file

*/

#define LAYER(i) (*layer+(i))

void read_data(struct lay **layer,int *nz,FILE *fp_in,my_boolean is_three_d,
	       char *layerfile)
{
  int i,j,k=0,n_expected_fields,n_fields,*n_alloc,loc_nz,
    icheck_val,ilayer;
  double *in;
  char	buffer[BUFSIZ],format[200];
  float dz1,dz2,dzmean=0.0;
  float tolerance = 0.05;	/* tolerance for depth layers */
  FILE *ifile;
  *layer = (struct lay *)GMT_memory (VNULL, (size_t)1, sizeof(struct lay), 
				     GMT_program);
  if(is_three_d){
    /* 
       read in the depth layers 
    */
    ifile = fopen(layerfile,"r");
    if(!ifile){
      fprintf(stderr, "read_data: can not open layer file %s\n",
	      layerfile);
      
      exit (EXIT_FAILURE);
    }
    dz1 = dz2 = -1.0;
    dzmean = 0.0;
    loc_nz = 0;
    while(fscanf(ifile,"%f",&(LAYER(loc_nz)->z) ) == 1){ /* read in each depth layer */
      /* 
	 add one more layer pointer
      */
      *layer = (struct lay *) GMT_memory ((void *)*layer, (size_t)(loc_nz+2), 
					  sizeof(struct lay), GMT_program);
      if(loc_nz > 0){		/* check for increasing layers */
	if(dz1 < 0){	
	  /* init first interval */
	  dz1 = LAYER(loc_nz)->z - LAYER(loc_nz-1)->z;
	  dz2 = dz1;
	}else{
	  /* later intervals */
	  dz2 = LAYER(loc_nz)->z - LAYER(loc_nz-1)->z;
	}
	if(dz2 <= 0.0){		/* check for monotonic increase */
	  fprintf(stderr,"read_data: error: levels in %s have to increase monotonically: n: %i dz; %g\n",
		  layerfile,loc_nz,dz2);
	  exit(EXIT_FAILURE);
	}
	dzmean += dz2;
	if(loc_nz > 1){		/* check for constant interval */
	  if(fabs(dz1-dz2)/dz1 > tolerance){ 
	    /* allow for some slop in the vertical spacing */
	    fprintf(stderr,"read_data: error: levels in %s have to be evenly spaced: n: %i dz(1): %g dz(n): %g\n",
		    layerfile,loc_nz,dz1,dz2);
	    fprintf(stderr,"read_data: tolerance: %g deviation: %g\n",tolerance,
		    fabs(dz1-dz2)/dz1);
	    exit(EXIT_FAILURE);
	  }
	}
      }
      loc_nz ++;
    }
    fclose(ifile);
    *layer = (struct lay *) 
      GMT_memory ((void *)*layer, (size_t)loc_nz, 
		  sizeof(struct lay), GMT_program);
    /* 
       re-assign perfectly spaced layers 
    */
    dzmean /= (float)(loc_nz-1);
    j = loc_nz - 1;
    dz1 = LAYER(0)->z + dzmean;
    for(i=1;i < j;i++,dz1 += dzmean){
      LAYER(i)->z = dz1;
    }
    if(gmtdefs.verbose){
      fprintf(stderr,"read_data: read %i layers from %s: zmin: %g dz: %g zmax: %g\n",
	      loc_nz,layerfile,LAYER(0)->z,dzmean,LAYER(loc_nz-1)->z);
    }
    /* end three 3 loop */
  }else{
    loc_nz = 1;
    LAYER(0)->z=0;
  }
  n_alloc = (int *) GMT_memory (VNULL, (size_t)(loc_nz), sizeof(int), GMT_program);
  for(i=0;i < loc_nz;i++){
    n_alloc[i] = GMT_CHUNK;
    /* init the layer structure */
    LAYER(i)->n= LAYER(i)->nmin = LAYER(i)->nmax = 0;
    LAYER(i)->lon = (float *) GMT_memory (VNULL, (size_t)n_alloc[i], sizeof(float), GMT_program);
    LAYER(i)->lat = (float *) GMT_memory (VNULL, (size_t)n_alloc[i], sizeof(float), GMT_program);
    LAYER(i)->f = (float *) GMT_memory (VNULL, (size_t)n_alloc[i], sizeof(float), GMT_program);
    LAYER(i)->fmean = 0.0;
    LAYER(i)->fmin =  FLT_MAX;
    LAYER(i)->fmax = -FLT_MAX;
  }
  /* 

  select the number of expected fields 

  */
  if(GMT_io.binary[0]){
    n_expected_fields = GMT_io.ncol[0] ;
  }else{
    n_expected_fields = (is_three_d)?(4):(3);
  }
  /* 
     which field holds the data? 
  */
  icheck_val = (is_three_d)?(3):(2);
  /* 
     read in possible header lines
  */
  if (!GMT_io.binary[0] && gmtdefs.io_header) 
    for (i = 0; i < gmtdefs.n_header_recs; i++) 
      GMT_fgets (buffer, BUFSIZ, fp_in);
  /* 

  start the main data input loop

  */
  while ((n_fields = GMT_input (fp_in, &n_expected_fields, &in)) >= 0 
	 && !(GMT_io.status & GMT_IO_EOF)) {	/* Not yet EOF */
    if (GMT_io.status & GMT_IO_MISMATCH) {
      fprintf (stderr, "%s: Mismatch between actual (%d) and expected (%d) fields near line %d\n", 
	       GMT_program, n_fields, n_expected_fields, k);
      exit (EXIT_FAILURE);
    }
    /* 
       test for NaN, if Nan, skip this entry
    */
    if (GMT_is_dnan (in[icheck_val])) 
      continue;
    /* 
       decide on which layer
    */
    if(is_three_d){
      /* 
	 have to find a layer, depth was specified
      */
      ilayer = (int)((float)((in[2]- LAYER(0)->z)/dzmean)+.5);
      if((ilayer < 0)||(ilayer >= loc_nz)){
	fprintf(stderr,"%s: error: datapoint at lon: %g lat: %g z: %g is out of depth range (%g - %g)\n",
		GMT_program,in[0],in[1],in[2],LAYER(0)->z,LAYER(loc_nz-1)->z);
	exit(EXIT_FAILURE);
      }
    }else{
      ilayer = 0;
    }
    /* 
       
    add entry to layer
    
    */
    LAYER(ilayer)->lon[LAYER(ilayer)->n] = (float)in[0];
    LAYER(ilayer)->lat[LAYER(ilayer)->n] = (float)in[1];
    LAYER(ilayer)->f[LAYER(ilayer)->n] = (float)in[icheck_val];
    /* find extrema */
    if (LAYER(ilayer)->fmin > in[icheck_val]){
      LAYER(ilayer)->fmin = (float)in[icheck_val];
      LAYER(ilayer)->nmin = (float)LAYER(ilayer)->n;
    }
    if (LAYER(ilayer)->fmax < in[icheck_val]) {
      LAYER(ilayer)->fmax = (float)in[icheck_val];
      LAYER(ilayer)->nmax = (float)LAYER(ilayer)->n;
    }
    /* increment counter */  
    LAYER(ilayer)->n++;
    LAYER(ilayer)->fmean += in[icheck_val]; /* add to mean */
    if (LAYER(ilayer)->n == n_alloc[ilayer]) { /* grow pointers */
      n_alloc[ilayer] += GMT_CHUNK;
      LAYER(ilayer)->lon = (float *) 
	GMT_memory ((void *)LAYER(ilayer)->lon, (size_t)n_alloc[ilayer], sizeof(float), GMT_program);
      LAYER(ilayer)->lat = (float *) 
	GMT_memory ((void *)LAYER(ilayer)->lat, (size_t)n_alloc[ilayer], sizeof(float), GMT_program);
      LAYER(ilayer)->f = (float *) 
	GMT_memory ((void *)LAYER(ilayer)->f, (size_t)n_alloc[ilayer], sizeof(float), GMT_program);
    }
  } /* end of input loop */
  if (fp_in != GMT_stdin) 
    GMT_fclose (fp_in);
  
  for(ilayer=0;ilayer < loc_nz;ilayer++){
    /* 
       wrap up and check all layers 
    */
    if (LAYER(ilayer)->n == 0) {
      if(loc_nz == 1)
	fprintf (stderr, "%s:  No datapoints inside region, aborts\n", GMT_program);
      else
	fprintf (stderr, "%s:  No datapoints inside layer %i (z: %g), aborts\n", 
		 GMT_program,ilayer+1,LAYER(ilayer)->z);
      exit (EXIT_FAILURE);
    }else{
      if(gmtdefs.verbose){
	fprintf (stderr, "%s: read %i datapoints ", 
		 GMT_program,LAYER(ilayer)->n);
	if(loc_nz > 1)
	  fprintf(stderr,"  for layer %i at depth: %g\n",
		  ilayer+1,LAYER(ilayer)->z);
	else
	  fprintf(stderr,"\n");
      }
    }
    /* average */
    LAYER(ilayer)->fmean /= (float)(LAYER(ilayer)->n);
    if (gmtdefs.verbose) {
      sprintf(format, "%s %s %s\n", gmtdefs.d_format, gmtdefs.d_format, gmtdefs.d_format);
      fprintf(stderr, "%s: Minimum value of your dataset x,y,z at: ", GMT_program);
      fprintf(stderr, format, (double)(LAYER(ilayer)->lon[LAYER(ilayer)->nmin]), 
	      (double)(LAYER(ilayer)->lat[LAYER(ilayer)->nmin]),
	      (double)(LAYER(ilayer)->fmin));
      fprintf(stderr, "%s: Maximum value of your dataset x,y,z at: ", GMT_program);
      fprintf(stderr, format, (double)(LAYER(ilayer)->lon[LAYER(ilayer)->nmax]), 
	      (double)(LAYER(ilayer)->lat[LAYER(ilayer)->nmax]), 
	      (double)(LAYER(ilayer)->fmax));
    }
    LAYER(ilayer)->lon = (float *) GMT_memory ((void *)(LAYER(ilayer)->lon), (size_t)(LAYER(ilayer)->n), sizeof(float), GMT_program);
    LAYER(ilayer)->lat = (float *) GMT_memory ((void *)(LAYER(ilayer)->lat), (size_t)(LAYER(ilayer)->n), sizeof(float), GMT_program);
    LAYER(ilayer)->f = (float *) GMT_memory ((void *)(LAYER(ilayer)->f), (size_t)(LAYER(ilayer)->n), sizeof(float), GMT_program);
  } /* end ilayer loop */
  *nz = loc_nz;
  free(n_alloc);
}
#undef LAYER
  



/* 

those are miscellaneous functions ssrf_driver is using

*/
/*
  
allocate floating point vector

*/
void my_vecalloc(float **x,int n,char *message)
{
  *x = (float *)malloc(sizeof(float)*(size_t)n);
  if(! (*x))
    MEMERROR(message);
}
/*
  
allocate floating point vector

*/
void my_ivecalloc(int **ix,int n,char *message)
{
  *ix = (int *)malloc(sizeof(int)*(size_t)n);
  if(! (*ix))
    MEMERROR(message);
}
/* reallocate floating point vector */
void my_vecrealloc(float **x,int n,char *message)
{
  *x = (float *)realloc(*x,
			sizeof(float)*(size_t)n);
  if(!(*x))
    MEMERROR(message);
}
/* reallocate int point vector */
void my_ivecrealloc(int **ix,int n,char *message)
{
  *ix = (int *)realloc(*ix,sizeof(int)*(size_t)n);
  if(!(*ix))
    MEMERROR(message);
}

void scale_vector(float *x,float fac,int n)
{
  int i;
  for(i=0;i<n;i++)
    x[i] *= fac;
}
// b = a 
void copy_vector(float *a, float *b, int n)
{
  memcpy(b,a,sizeof(float)*n);
}

