#include "ssrf.h"
/*


miscellaneous linear algebra type routines


$Id: linalg.c,v 1.3 2004/02/25 00:05:49 becker Exp $



*/

/*

low-level, BLAS type calculations


TO-DO: change these to memcpy statements

*/
COMP_PRECISION vector_diff(COMP_PRECISION *x,COMP_PRECISION *y,
			   int n)
{
  int i;
  COMP_PRECISION sum=0.0,tmp;
  for(i=0;i<n;i++){
    tmp = x[i]-y[i];
    sum += tmp * tmp;
  }
  if(sum>0)return sqrt(sum);else return 0.0;
}
void zero_vector(COMP_PRECISION *x,int n)
{
  int i;
  for(i=0;i<n;i++)
    x[i]=0.0;
}
void scale_vector(COMP_PRECISION *x,COMP_PRECISION fac,int n)
{
  int i;
  for(i=0;i<n;i++)
    x[i] *= fac;
}
void scale_vector3d(COMP_PRECISION *x,COMP_PRECISION fac)
{
  int i;
  for(i=0;i<3;i++)
    x[i] *= fac;
}
// b = a 
void copy_vector(COMP_PRECISION *a, COMP_PRECISION *b, int n)
{
  memcpy(b,a,sizeof(COMP_PRECISION)*n);
}
void copy_3dvector(COMP_PRECISION *a, COMP_PRECISION *b)
{
  memcpy(b,a,sizeof(COMP_PRECISION)*3);
}
void copy_3dmatrix(COMP_PRECISION *a, COMP_PRECISION *b)
{
  memcpy(b,a,sizeof(COMP_PRECISION)*9);
}
void add_a_to_b_vector(COMP_PRECISION *a, COMP_PRECISION *b,
		       int n)
{
  int i;
  for(i=0;i<n;i++)
    b[i] += a[i];
}
// 3-D vector dot product
COMP_PRECISION vec3ddotp(COMP_PRECISION *a,COMP_PRECISION *b)
{
  int i;
  COMP_PRECISION tmp=0.0;
  for(i=0;i < 3;i++)
    tmp += a[i] * b[i];
  return tmp;
}
COMP_PRECISION norm3d(COMP_PRECISION *a)
{
  COMP_PRECISION xl;
  xl = vec3ddotp(a,a);
  if(xl <= 0.0)
    return 0.0;
  else
    return sqrt(xl);
}

/*

find the largest element of a vector

*/
COMP_PRECISION max_vec(COMP_PRECISION *x, int n)
{
  int i;
  COMP_PRECISION max=-1e30;
  for(i=0;i<n;i++)
    if(x[i] > max)
      max = x[i];
  return max;
}

/*

find the maximum of the absolute values of a vector

*/

COMP_PRECISION max_abs_vec(COMP_PRECISION *x, int n)
{
  int i;
  COMP_PRECISION max=0.0,tmp;
  for(i=0;i<n;i++){
    tmp = fabs(x[i]);
    if(tmp > max)
      max = tmp;
  }
  return max;
}

//  \vec{a} = \vec{b}
void a_equals_b_vector(COMP_PRECISION *a,COMP_PRECISION *b,
		       int n)
{
  copy_vector(b,a,n);
}
/* 
   
given a x[n][n] C style matrix, return the unity matrix

*/
void unity_matrix(COMP_PRECISION *x,int n)
{
  int i,j;
  for(i=0;i<n;i++)
    for(j=0;j<n;j++)
      x[i*n+j] = (i==j)?(1.0):(0.0);
}
