/* 

header file for SSRFPACK (by Renka) wrappers

Thorsten Becker (thwbecker@post.harvard.edu)

$Id: ssrf.h,v 1.3 2004/03/08 03:16:50 becker Exp becker $

 */
#include <stdio.h>
#include <math.h>
#include <limits.h>
#include <stdlib.h>
#include <string.h>

#ifdef SSRF_SINGLE_PRECISION 
#define COMP_PRECISION float
#define EPS_COMP_PREC 5e-7
#else
#define COMP_PRECISION double
#define EPS_COMP_PREC 5e-15
#endif

#define my_boolean unsigned char
#define TRUE 1
#define FALSE 0


#ifndef MEMERROR
#define MEMERROR(x) {fprintf(stderr,"memory allocation error: %s\n",x);exit(-1);}
#endif
#ifndef PE
#define PE(x) {fprintf(stderr,"%s\n",x);}
#endif

/* 

   interpolation structures

*/
struct ssrf_par{
  COMP_PRECISION *x,*y,*z;	/* cartesian coordinates */
  int *list,*lptr,*lend;	/* for triangulation */
  int istart;			/* start node for search */
  int lnew;	
  /* variable tension flag */
  my_boolean variable_tension,cart_coord_init,triangulation_init,
    sigma_init;
  COMP_PRECISION *sigma;	
  /* tolerances */
  COMP_PRECISION getsig_tol,getgrad_tol;
  int imode;			/* interpolation mode */
  int maxit, itgs;

};

/* C prototypes */
#ifdef SSRF_SINGLE_PRECISION
#include "auto_proto.sgl.h"
#else
#include "auto_proto.h"
#endif
/* 

FORTRAN functions

*/
#ifdef SGI_SUBROUTINE_CONVENTION
#define trans trans_
#define trmesh trmesh_
#define intrc0 intrc0_
#define intrc1 intrc1_
#define gradl gradl_
#define gradg gradg_
#define getsig getsig_
#ifdef GCC_USCR
#define ssrf_norm3d ssrf_norm3d__
#define ssrf_snhcsh ssrf_snhcsh__
#define ssrf_hypot ssrf_hypot__
#else
#define ssrf_norm3d ssrf_norm3d_
#define ssrf_snhcsh ssrf_snhcsh_
#define ssrf_hypot ssrf_hypot_
#endif
#endif

extern void trans(int *, COMP_PRECISION *,COMP_PRECISION *,COMP_PRECISION *,
		  COMP_PRECISION *,COMP_PRECISION *);

extern void trmesh(int *, COMP_PRECISION *,COMP_PRECISION *,COMP_PRECISION *,
		   int *, int *, int *,int *, int *,int *,
		   COMP_PRECISION *, int *);

extern void intrc0(int *, COMP_PRECISION *,COMP_PRECISION *,COMP_PRECISION *,
		   COMP_PRECISION *,COMP_PRECISION *,COMP_PRECISION *,
		   int *, int *, int *, int *,
		   COMP_PRECISION *, int *);
extern void intrc1(int *, COMP_PRECISION *,COMP_PRECISION *,
		   COMP_PRECISION *,COMP_PRECISION *,COMP_PRECISION *,COMP_PRECISION *,
		   int *, int *, int *, 
		   int *, COMP_PRECISION *, int *,COMP_PRECISION *, int *,
		   COMP_PRECISION *,int *);

extern void gradl(int *,int *,COMP_PRECISION *,COMP_PRECISION *,COMP_PRECISION *,
		  COMP_PRECISION *,int *,int *,int *,COMP_PRECISION *,
		  int *);

extern void gradg(int *,COMP_PRECISION *,COMP_PRECISION *,COMP_PRECISION *,
		  COMP_PRECISION *,int *,int *,int *,int *,COMP_PRECISION *,
		  int *,COMP_PRECISION *,COMP_PRECISION *,int *);

extern void getsig(int *,COMP_PRECISION *,COMP_PRECISION *,COMP_PRECISION *,
		   COMP_PRECISION *,int *, int *, int *,COMP_PRECISION *,
		   COMP_PRECISION *,COMP_PRECISION *,COMP_PRECISION *,
		   int *);
